#include <stdlib.h>             // Needed for "atoi".

#include <string>
using std::string;

#include "..\inet\Socket.h"
#include "..\escutil\ESCTime.h"
#include "..\BoardDLLs\brd_generic.h"
#include "..\escutil\iv_str.h"   // FindCaseInsensitiveString
#include "step_audioplay.h"      // StepAudioPlay
#include "..\TFEngine\TFEngine_communication.h"
#include "step_HTTP.h"

StepHTTP::StepHTTP(STEPREC& stepRecord) : Step(stepRecord)
{
   requestBody = stepRecord.s9.arr;
   responseBodyVar = stepRecord.s1.arr;
   responseHeaderVar = stepRecord.s8.arr;
   timeout = stepRecord.s2.arr;
   URL = stepRecord.s6.arr;
   msgCd = stepRecord.s3.arr;
   fileNm = stepRecord.s4.arr;
   connectionTimeout = stepRecord.s5.arr;
   connectionRetries = stepRecord.s7.arr;
   format = (SOUND_FORMAT) stepRecord.u1;
   requestMethod = (HTTPRequest::REQUEST_METHOD) stepRecord.u2;
   certStoreOption = (TCPSocket::SSL_CERT_STORE_OPT) stepRecord.u3;
}

StepHTTP::StepHTTP(void) : Step(uninitialisedStepRecord)
{
#ifdef __BORLANDC__
   certStoreOption = TCPSocket::STORE_NONE;
#else
   certStoreOption = TCPSocket::SSL_CERT_STORE_OPT::STORE_NONE;
#endif
}

StepHTTP::~StepHTTP(void)
{
}

const char* StepHTTP::GetStepName(void) const
{
	return("HTTP");
}

void StepHTTP::SetRequestBody(const char* value)
{
   requestBody = value == NULL ? "" : value;
}

void StepHTTP::SetRequestType(HTTPRequest::REQUEST_METHOD value)
{
   requestMethod = value;
}

void StepHTTP::SetResponseBodyVar(const char* value)
{
   responseBodyVar = value == NULL ? "" : value;
}

void StepHTTP::SetResponseHeaderVar(const char* value)
{
   responseHeaderVar = value == NULL ? "" : value;
}

void StepHTTP::SetTimeout(const char* value)
{
   timeout = value == NULL ? "" : value;
}

void StepHTTP::SetURL(const char* value)
{
   URL = value == NULL ? "" : value;
}

void StepHTTP::SetMessageCode(const char* value)
{
   msgCd = value == NULL ? "" : value;
}

void  StepHTTP::SetFileName(const char* value)
{
   fileNm = value == NULL ? "" : value;
}

void StepHTTP::SetFileType(int value)
{
   format = (SOUND_FORMAT) value;
}

void StepHTTP::SetConnectionTimeout(const char* value)
{
   connectionTimeout = value == NULL ? "" : value;
}

void StepHTTP::SetConnectionRetries(const char* value)
{
   connectionRetries = value == NULL ? "" : value;
}

IV_STATUS StepHTTP::Execute(PIVRINFO pInfo)
{
   IV_STATUS status;
   char* value = NULL;
   // Evaluate all parameters before exiting
   bool bParmsValid = true;
   string fileName;
   int iTimeout;
   int iConnectionTimeout;
   int iConnectionRetries;
   HTTPTransaction http(pInfo->ivLog);
   unsigned int iContentLength;

   if(Status(pInfo) != IVR_SUCCESS)
   {
      return(IVR_FAIL);
   }

   // --- Retrieve timeout pszTimeoutString (in milliseconds)
   if((value = pInfo->variables->ParseParameter("Silence TIMEOUT", timeout.c_str(), "@TIMEOUT")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      iTimeout = atoi(value);
      delete[] value;

      if((iTimeout < 0) && (iTimeout != FOREVER))
      {
		   pInfo->ivLog->Print("      ERROR: Silence TIMEOUT '%s' must be >= 0\n",
            pInfo->variables->IsParsedEmpty(timeout.c_str()) ? "@TIMEOUT" : timeout.c_str());
         bParmsValid = false;
      }
   }

   // Retrieve connection timeout
   if((value = pInfo->variables->ParseParameter("HTTP Initial Connection TIMEOUT", connectionTimeout.c_str(), "500")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      iConnectionTimeout = atoi(value);
      delete[] value;

      if((iConnectionTimeout <= 0) && (iTimeout != FOREVER))
      {
		   pInfo->ivLog->Print("      ERROR: HTTP Initial Connection TIMEOUT '%s' must be > 0\n", connectionTimeout.c_str());
         bParmsValid = false;
      }
   }

   // Retrieve connection retries
   if((value = pInfo->variables->ParseParameter("HTTP Initial Connection Retries", connectionRetries.c_str(), "2")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      iConnectionRetries = atoi(value);
      delete[] value;

      if(iConnectionRetries < 0)
      {
		   pInfo->ivLog->Print("      ERROR: HTTP Initial Connection Retries '%s' must be >= 0\n", connectionRetries.c_str());
         bParmsValid = false;
      }
   }

   // --- Retrieve filename
   // Message code
   if(!msgCd.empty())
   {
      if((value = pInfo->variables->ParseParameter("Script code", msgCd.c_str())) == NULL)
      {
         bParmsValid = false;
      }
      else
      {
         char* scriptCodeFileName;

         status = GetScriptCodeFileNameAndFormat(pInfo, value, scriptCodeFileName, format);
         delete[] value;

         if(status != IVR_SUCCESS)
         {
            return(status);
         }

         fileName = scriptCodeFileName;
         delete[] scriptCodeFileName;
      }
   }
   // File name
   else if(!fileNm.empty())
   {
      if((value = pInfo->variables->ParseParameter("File name", fileNm.c_str())) == NULL)
      {
         bParmsValid = false;
      }
      else
      {
         fileName = value;
         delete[] value;
      }
   }

   // --- Retrieve URL
   if((value = pInfo->variables->ParseParameter("URL", URL.c_str())) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      http.Request()->SetMethod(requestMethod);
      http.SetURL(value);
      delete[] value;
   }

   // --- Certificate Store Option
   http.set_CertStoreOpt(certStoreOption);
   pInfo->ivLog->Print("      SSL Certificate store option is '%s'\n", http.get_CertStoreOpt());

#ifdef __BORLANDC__
   // ACE Library has dropped support for Borland compiler
   // We cannot implement the SSL Certificate store comparisons change unless we migrate
   // TeleFlow 32-bit compile to Microsoft Visual Studio
   if(certStoreOption == TCPSocket::WINDOWS_ROOT)
   {
      pInfo->ivLog->Print("         Serching certificate stores is not available in this version of TeleFlow\n");
      return(IVR_NOTSUPPORTED);
   }
#endif

   // --- Retrieve HTTP Request Message Body
   if((value = pInfo->variables->ParseParameter("HTTP Request Message Body", requestBody.c_str(), "")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      http.Request()->SetBody(value);
      delete[] value;
   }

   // Initialise value to empty string
   pInfo->variables->Set(responseBodyVar.c_str(), "");

   if(!bParmsValid)
   {
      return(IVR_FATAL);
   }

   // Plays a file
   if(!fileName.empty())
   {
      char* pathName;

   	if(PortRequired(pInfo) != IVR_SUCCESS)
      {
         return(IVR_FATAL);
      }

      pathName = pInfo->variables->FilePath(fileName.c_str());

      status = pInfo->ivBoard->HTTPRequest(http, iTimeout, iConnectionTimeout,
         iConnectionRetries, pathName, format, atoi(pInfo->variables->Get("@VOLUME")));

      delete[] pathName;
   }
   // No file played
   else
   {
      LARGE_INTEGER liTimeout;
      ESCTime startTime;
      ESCTime currentTime;
      HANDLE waitObjs[3];
      HANDLE hTimer = CreateWaitableTimer(0, false, 0);
      int ret;
      bool loopAgain;
      bool attemptConnect;
      int connectCounter = 0;

      status = IVR_SUCCESS;

      startTime.Update();

      // Set up events array
      waitObjs[0] = pInfo->messages->GetMessageSignal();  // system event such as shutdown
      waitObjs[1] = hTimer;                               // Silence timeout
      waitObjs[2] = http.GetEventHandle();       // signal from socket

   	// Set timeout timer
      if(iConnectionTimeout != FOREVER)
      {
         if(iTimeout < iConnectionTimeout)
         {
            liTimeout.QuadPart = -(LONGLONG) iTimeout * 10000;
         }
         else
         {
            liTimeout.QuadPart = -(LONGLONG) iConnectionTimeout * 10000;
         }

         SetWaitableTimer(hTimer, &liTimeout, 0, 0, 0, false);
      }
      else if(iTimeout != FOREVER)
      {
         liTimeout.QuadPart = -(LONGLONG) iTimeout * 10000;
         SetWaitableTimer(hTimer, &liTimeout, 0, 0, 0, false);
      }

      // Try up to iConnectionRetries times to connect
      loopAgain = true;
      attemptConnect = true;
      while(loopAgain)
      {
         // Connect to HTTP server
         if(attemptConnect)
         {
            pInfo->ivLog->Print("      Connection attempt %d\n", ++connectCounter);

            if((ret = http.Connect()) != Socket::SUCCESS)
            {
               const char* errorMessage = Socket::WSAErrorString(ret);

               pInfo->ivLog->Print("      ERROR: Unable to connect to host\n");
               pInfo->ivLog->Print("         '%s'\n", errorMessage);
               if(!responseHeaderVar.empty())
               {
                  pInfo->variables->Set(responseHeaderVar.c_str(), errorMessage);
               }
               CloseHandle(hTimer);
               return(IVR_FAIL);
            }
         }

         switch(WaitForMultipleObjects(3, waitObjs, false, INFINITE))
         {
            // System event
            case WAIT_OBJECT_0:
            {
               status = pInfo->ivBoard->ProcessSystemEvent();

               if(status != IVR_SUCCESS)
               {
                  CloseHandle(hTimer);
                  return(status);
               }

               attemptConnect = false;
               break;
            }

            // Timeout
            case WAIT_OBJECT_0 + 1:
            {
               int updatedTimeout;

               pInfo->ivLog->Print("      ERROR: Timed out trying to connect to host\n");

               if(iTimeout != FOREVER)
               {
                  currentTime.Update();
                  updatedTimeout = iTimeout - (int) (1000*currentTime.GetSecondsSince(startTime));
               }
               else
               {
                  updatedTimeout = iConnectionTimeout;
               }

               // Silence timeout or maximum connection attempts
               if((updatedTimeout <= 0) || (connectCounter > iConnectionRetries))
               {
                  CloseHandle(hTimer);
                  return(IVR_TIMEOUT);
               }
         	   // Reset timeout timer
               else
               {
                  liTimeout.QuadPart = -(LONGLONG) (updatedTimeout < iConnectionTimeout ? updatedTimeout : iConnectionTimeout) * 10000;
                  SetWaitableTimer(hTimer, &liTimeout, 0, 0, 0, false);
               }

               http.Close();
               attemptConnect = true;
               break;
            }

            // Connection signalled
            case WAIT_OBJECT_0 + 2:
            {
               // Check status
               if((ret = http.GetStatus()) == Socket::SUCCESS)
               {
                  pInfo->ivLog->Print("      Connection successful\n");
                  http.SelectEvents(FD_READ | FD_CLOSE);
                  loopAgain = false;
               }
               else
               {
                  const char* errorMessage = Socket::WSAErrorString(ret);

                  pInfo->ivLog->Print("      Connection attempt failed\n");
                  pInfo->ivLog->Print("         '%s'\n", errorMessage);

                  if(connectCounter > iConnectionRetries)
                  {
                     if(!responseHeaderVar.empty())
                     {
                        pInfo->variables->Set(responseHeaderVar.c_str(), errorMessage);
                     }
                     return(IVR_FAIL);
                  }

                  http.Close();
                  attemptConnect = true;
               }

               break;
            }

            // WaitForMultipleObjects failed
            case WAIT_FAILED:
            {
               DWORD errorID = GetLastError();
               char MsgBuf[1024];

               FormatMessage(
                  FORMAT_MESSAGE_FROM_SYSTEM,
                  0,
                  errorID,
                  MAKELANGID (LANG_NEUTRAL, SUBLANG_DEFAULT),
                  MsgBuf,
                  1024,
                  0
               );

               pInfo->ivLog->Print("      ERROR: WaitForMultipleObjects returned: %x: %s", errorID, MsgBuf);
               if(!responseHeaderVar.empty())
               {
                  pInfo->variables->Set(responseHeaderVar.c_str(), MsgBuf);
               }
               CloseHandle(hTimer);
               return(IVR_FAIL);
            }
         }
      }

   	// Set silence timeout timer
      if(iTimeout != FOREVER)
      {
         currentTime.Update();
         liTimeout.QuadPart = -(LONGLONG) (iTimeout - (int) currentTime.GetSecondsSince(startTime)) * 10000;
         SetWaitableTimer(hTimer, &liTimeout, 0, 0, 0, false);
      }

      if(http.IsSSLEnabled())
      {
         pInfo->ivLog->Print("      Start SSL handshake\n");
         if((ret = http.SSLHandshake()) == Socket::SUCCESS)
         {
            pInfo->ivLog->Print("      SSL Handshake successful\n");
         }
         else
         {
            const char* errorMessage = Socket::WSAErrorString(ret);

            pInfo->ivLog->Print("      SSL handshake failed\n");
            pInfo->ivLog->Print("         %s\n", errorMessage);
            if(!responseHeaderVar.empty())
            {
               pInfo->variables->Set(responseHeaderVar.c_str(), errorMessage);
            }
            CloseHandle(hTimer);
            return(IVR_FAIL);
         }
      }

      pInfo->ivLog->Print("      Sending request: '%s'\n", http.Request()->GetText());

      // Send the request
      while((ret = http.Send()) == WSAEWOULDBLOCK)
      {
         // Allow 10 ms for socket to clear
         Sleep(10);
      }

      if(ret != Socket::SUCCESS)
      {
         const char* errorMessage = Socket::WSAErrorString(ret);

         pInfo->ivLog->Print("      ERROR: Failed to send request\n");
         pInfo->ivLog->Print("         '%s'\n", errorMessage);
         if(!responseHeaderVar.empty())
         {
            pInfo->variables->Set(responseHeaderVar.c_str(), errorMessage);
         }
         CloseHandle(hTimer);
         return(IVR_FAIL);
      }

      pInfo->ivLog->Print("      Request successful, waiting for response\n");

      loopAgain = true;
      pInfo->ivLog->Print("      @@@-debug, loop-1.\n");
      while(loopAgain)
      {
         pInfo->ivLog->Print("      @@@-debug, loop-1: Wait for multiple objects.\n");
         switch(WaitForMultipleObjects(3, waitObjs, false, INFINITE))
         {
            // System event
            case WAIT_OBJECT_0:
            {
               pInfo->ivLog->Print("      @@@-debug, loop-1: system event received.\n");
               status = pInfo->ivBoard->ProcessSystemEvent();

               if(status != IVR_SUCCESS)
               {
                  CloseHandle(hTimer);
                  return(status);
               }

               break;
            }

            // Timed out
            case WAIT_OBJECT_0 + 1:
            {
               pInfo->ivLog->Print("      ERROR: Timed out waiting for response\n");
               CloseHandle(hTimer);
               return(IVR_TIMEOUT);
            }

            // Socket signalled
            case WAIT_OBJECT_0 + 2:
            {
               pInfo->ivLog->Print("      @@@-debug, loop-1: Wait_Object_2, socket signalled.\n");
               while(loopAgain)
               {
                  if((ret = http.Receive()) == WSAEWOULDBLOCK)
                  {
                     // Allow 10 ms for socket to clear
                     pInfo->ivLog->Print("      @@@-debug, loop-1: WSAEWOULDBLOCK, wait 10 ms, try again.\n");
                     Sleep(10);
                  }
                  else if(ret != Socket::SUCCESS)
                  {
                     const char* errorMessage = Socket::WSAErrorString(ret);

                     pInfo->ivLog->Print("      ERROR: Failed to receive response\n");
                     pInfo->ivLog->Print("         '%s'\n", errorMessage);
                     if(!responseHeaderVar.empty())
                     {
                        pInfo->variables->Set(responseHeaderVar.c_str(), errorMessage);
                     }
                     CloseHandle(hTimer);
                     return(IVR_FAIL);
                  }

                  if(strlen(http.Response()->GetText()) == 0)
                  {
                     pInfo->ivLog->Print("      @@@-debug, loop-1: GetText() empty (length is 0).\n");
                     Sleep(10);
                  }
                  else
                  {
                     pInfo->ivLog->Print("      @@@-debug, loop-1: GetText() contains data, exit initial loop\n");
                     loopAgain = false;
                  }
               }

               break;
            }

            // WaitForMultipleObjects failed
            case WAIT_FAILED:
            {
               DWORD errorID = GetLastError();
               char MsgBuf[1024];

               FormatMessage(
                  FORMAT_MESSAGE_FROM_SYSTEM,
                  0,
                  errorID,
                  MAKELANGID (LANG_NEUTRAL, SUBLANG_DEFAULT),
                  MsgBuf,
                  1024,
                  0
               );

               pInfo->ivLog->Print("      ERROR: WaitForMultipleObjects returned: %x: %s", errorID, MsgBuf);
               if(!responseHeaderVar.empty())
               {
                  pInfo->variables->Set(responseHeaderVar.c_str(), MsgBuf);
               }
               CloseHandle(hTimer);
               return(IVR_FAIL);
            }
         }
      }

      pInfo->ivLog->Print("      Response status code: %u\n", http.Response()->GetStatusCode());

      // Check that response was OK (20x)
      // Bad return code: not 200 to 209
      if((http.Response()->GetStatusCode() < 200) || (http.Response()->GetStatusCode() > 209))
      {
         pInfo->ivLog->Print("      ERROR: '%s'\n", http.Response()->GetText());
         CloseHandle(hTimer);
         return(IVR_FAIL);
      }

      // Don't have whole message yet
      loopAgain = true;
      pInfo->ivLog->Print("      @@@-debug, loop-2.\n");
      while((!http.Response()->HaveCompleteMessage(&iContentLength)) && (loopAgain))
      {
         pInfo->ivLog->Print("      @@@-debug, loop-2: Don't have whole message yet\n");
         switch(WaitForMultipleObjects(3, waitObjs, false, INFINITE))
         {
            // System event
            case WAIT_OBJECT_0:
            {
               pInfo->ivLog->Print("      @@@-debug, loop-2: system event received.\n");
               status = pInfo->ivBoard->ProcessSystemEvent();

               if(status != IVR_SUCCESS)
               {
                  CloseHandle(hTimer);
                  return(status);
               }

               break;
            }

            // Timed out
            case WAIT_OBJECT_0 + 1:
            {
               pInfo->ivLog->Print("      ERROR: Timed out waiting for remainder of response\n");
               // Store as much of the message as we have received in the result variable
               pInfo->variables->Set(responseBodyVar.c_str(), http.Response()->GetBody());
               CloseHandle(hTimer);
               return(IVR_TIMEOUT);
            }

            // Socket signalled
            case WAIT_OBJECT_0 + 2:
            {
               if((ret = http.Receive()) == WSAEWOULDBLOCK)
               {
                  pInfo->ivLog->Print("      @@@-debug, loop-2: WSAEWOULDBLOCK, wait 10 ms, try again.\n");
                  // Allow 10 ms for socket to clear
                  Sleep(10);
               }
               // Allow HTTP to receive all data, even if header doesn't supply the size
               // Retrieve until we get a "connection reset" result
               else if((ret == WSAECONNRESET) && (strlen(http.Response()->GetText()) > 0))
               {
                  pInfo->ivLog->Print("      @@@-debug, loop-2: WSAECONNRESET, connection reset by peer.\n");
                  if(iContentLength == 0)
                  {
                     pInfo->ivLog->Print("      @@@-debug, loop-2: Exact length not defined; assume we are done.\n");
                     loopAgain = false;
                     break;
                  }
                  else
                  {
                     pInfo->ivLog->Print("      @@@-debug, loop-2: full message not received!\n");
                     CloseHandle(hTimer);
                     return(IVR_FAIL);
                  }
               }
               else if(ret != Socket::SUCCESS)
               {
                  const char* errorMessage = Socket::WSAErrorString(ret);

                  pInfo->ivLog->Print("      ERROR: Failed to receive remainder of response\n");
                  pInfo->ivLog->Print("         '%s'\n", errorMessage);
                  if(!responseHeaderVar.empty())
                  {
                     pInfo->variables->Set(responseHeaderVar.c_str(), errorMessage);
                  }
                  CloseHandle(hTimer);
                  return(IVR_FAIL);
               }

               break;
            }

            // WaitForMultipleObjects failed
            case WAIT_FAILED:
            {
               DWORD errorID = GetLastError();
               char MsgBuf[1024];

               FormatMessage(
                  FORMAT_MESSAGE_FROM_SYSTEM,
                  0,
                  errorID,
                  MAKELANGID (LANG_NEUTRAL, SUBLANG_DEFAULT),
                  MsgBuf,
                  1024,
                  0
               );

               pInfo->ivLog->Print("      ERROR: WaitForMultipleObjects returned: %x: %s", errorID, MsgBuf);
               if(!responseHeaderVar.empty())
               {
                  pInfo->variables->Set(responseHeaderVar.c_str(), MsgBuf);
               }
               CloseHandle(hTimer);
               return(IVR_FAIL);
            }
         }
      }

      pInfo->ivLog->Print("      Response received\n");

      CloseHandle(hTimer);

      status = IVR_SUCCESS;
   }

   if(responseHeaderVar.empty())
   {
      // Output header received in the log
      pInfo->ivLog->Print("      HTTP Response Header:\n");
      pInfo->ivLog->Print("         '%s'\n", http.Response()->GetHeader());
   }
   else if(!pInfo->variables->Set(responseHeaderVar.c_str(), http.Response()->GetHeader()))
   {
      pInfo->ivLog->Print("      ERROR: Failed to set HTTP Response Header variable '%s'\n", responseHeaderVar.c_str());
      return(IVR_FATAL);
   }

   if(!responseBodyVar.empty() && !pInfo->variables->Set(responseBodyVar.c_str(), http.Response()->GetBody()))
   {
      pInfo->ivLog->Print("      ERROR: Failed to set HTTP Response Body variable '%s'\n", responseBodyVar.c_str());
      return(IVR_FATAL);
   }

   return(status);
 }
/*StepHTTP::Execute*/
